/***************************************************************************
 *   Copyright (C) 2008 by Franz Reiter   *
 *   franz.reiter@cadcam.co.at   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


/*  APP_GIS1.c                                2008-01-24    Franz Reiter.

Linkjob: ../xa/APP_GIS1.mak


gCad_main            CAD-Core starts application here
gCad_fini            close application

gis_init_pl          get surf-pTab-List from DB
gis_init_UI          Create UI-panel (Toolbar)
gis_lst_sup          create list of sufaces with pTabs; display List.
gis_DrawPoints
gis_DrawPoint
gis_HideAllPoints
gis_HidePoints
gis_mdl_upd          create update / Modelcodes
gis_mdl_del_pTab     

gis_CB1              action-dispatcher
gis_sel_CB           CAD-Core reports selection
gis_Lst_CB           userselection of surfaces-listEntry
gis_UI_getPt         get coords from entryfields
gis_disp_surNr       display activeSurfNr & activePointNr
gis_disp_coords      display coords of active gis-point


gis_surPt_chg        change active point-surface
gis_ptab_load        user has pointfile selected; load it to pTab
gis_bload_pTab       load pTab from bin.file
gis_bsave_pTab       save pTab into bin.file
gis_PtabDel          delete (all Points from) surface ActSurNr
gis_freePos_pTab     get next free Position in pTab
gis_ptab_save        Export points into Ascii-file
gis_PtMod            modify Point Nr ActPtNr from surface ActSurNr
gis_PtDel            delete Point Nr ActPtNr from surface ActSurNr
gis_PtAdd            add Point to surface ActSurNr
gis_mesh             create/diplay mesh


gis_surTab_add
gis_surInd_surNr     get index into surTab from DB-index of pointSurface
gis_dump_surTab

ptab_file_load1      read Punkte aus einer Ascii-Datei nach pTab
ptab_cad_load1



=================================================================
*/
#ifdef globTag
 void gis(){}
#endif


#define VERSION " APP_GIS1 Version 1.00 / 2008-03-01"


#include "../xa/MS_Def1.h"

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>


#include <GL/gl.h>
#include <GL/glu.h>

#include <gdk/gdkkeysyms.h>            // GDK_Esc


// gtk muss vor __declspec(dllimport) stehen
#include "../gtk/ut_gtk.h"             // f. xa_ui.h


#define APP_NAME  "APP_GIS1"




#ifdef _MSC_VER
// die folgenden 2 Funktionen exportieren (werden vom Main gerufen):
__declspec(dllexport) int gCad_main ();
__declspec(dllexport) int gCad_fini ();
// nachfolgende externals werden aus dem Main-Exe imported:
#define extern __declspec(dllimport)
#endif




#include "../xa/ut_ui_TX.h"            // TX_Print
#include "../xa/xa_mem.h"              // memspc51
#include "../ut/ut_geo.h"              // Point ...
#include "../ut/ut_txt.h"              // fnam_del
#include "../gr/ut_UI.h"               // UI_Func... SYM_..
#include "../gr/ut_GL.h"               // GL_GetViewPos
#include "../db/ut_DB.h"               // DB_GetGTxt
#include "../gr/ut_DL.h"               // DB_GetDLatt



typedef struct {long indp, indm;
  unsigned visi:1, exi_p:1, exi_m:1, modif:1, unused:29;}  typ_sur_ptab;
// indp    DB-Index of PTAB-surf
// indm    DB-Index of MSH-surf
// exi_p   1: PTAB-surf-Code already exists in Model; 0=no.
// exi_m   1: MSH-surf-Code already exists in Model; 0=no.
// visi    0=not visible; 1=visible;
// modif   1=points modified; 0=not.



// ext aus xa.c:
extern char      WC_modnam[128];  // der Modelname
extern char      AP_dir_open[128];
extern  int      WC_modnr;        // Nr of the active submodel; -1=main.

// ex xa_ui.c:
extern GtkWidget  *UIw_Box_TB;    // toolbarBox


// // ex ../ut/ut_txfil.c
// extern char *UTF_FilBuf0;
// extern long UTF_FilBuf0Len;  // die aktuelle Filesize



// aus ../gr/ut_DL.c:
extern DL_Att  *GR_ObjTab;  // das ist die DL





//---------------- local vars: ---------------------
// a list of active surfaces from PTAB's
#define MAX_SURPTAB 50
static typ_sur_ptab surTab[MAX_SURPTAB];      // Liste der aktiven pTab-surfs
static int surTabNr;                          // nr of active pTab-surfs
static int surIndLast;


static int    ActSurNr,     // DB-Index (surTab.indp) of active point-surface
              ActPtNr,
              ActPtabNr, ActPtabModFlag;
static long   ActPtDli;     
static Point  *ActPtab = NULL;

static int    ActFtabNr, ActFtabModFlag;
static Fac3   *ActFtab = NULL;


static GtkWidget *tbGis1=NULL;    // GIS-Toolbar
static GtkWidget *wcbcp, *wcbdp, *wcbdi, *wlbas, *wlbap;
static GtkWidget *wenvx, *wenvy, *wenvz, *wls;
static GLfloat GL_col_tab [][4] = {
  0.7,  0.1,  0.1, 1.0,     //  0 col. points
  0.1,  0.1,  0.7, 1.0};    //  1 col. info
// red  green blue intens




//---------------- prototypes: ---------------------
  GtkWidget* gis_init_UI (GtkWidget *parent, void *data);
  int gis_sel_CB (int src, long dl_ind);
  int gis_CB1 (GtkWidget *parent, void *data);
  int gis_Lst_CB (GtkWidget *parent, void *data);
  int gis_ptab_load (char *fnam, char *dirNam);
  int gis_ptab_save (char *fnam, char *dirNam);
  int UFA_dump1 (Fac3 *fac, Fac3 *fnb, char *txt, ...);
  int gis_surInd_surNr (int surNr);




//=========================================================
  int gCad_main () {
//=========================================================
// user has selected this plugin; starting ...


  int     *iba, iNr;
  Mesh    ms1;

  // write to Main-Infowindow ..
  TX_Print("%s",VERSION);


  // create List of active pTabs
  // get list of surfaces with subTyp PTAB
  gis_init_pl ();   // get pTab-List from DB 



  // surNr = find last used surface-nr (A)
  surIndLast = DB_QueryLastUsed (Typ_SUR, 0);
    printf(" LastUsed A-nr = %d\n",surIndLast);


  // create & activate toolbar
  tbGis1 = gis_init_UI (UIw_Box_TB, NULL);

  ActSurNr = -1;
  if(surTabNr < 1) goto L_99;


  // update surface-list
  gis_lst_sup ();


  // activate last surf
  gis_surPt_chg (surTabNr-1);


  //----------------------------------------------------------------
  // attach grafic selections
  L_99:
  AP_UserSelection_get (gis_sel_CB);

  ActPtDli = -1;

  TX_Print("%d point-surfaces loaded",surTabNr);


  return 0;

}


//================================================================
  GtkWidget* gis_init_UI (GtkWidget *parent, void *data) {
//================================================================

  GtkWidget *tbc, *box0, *box1, *box2;

  printf("gis_init_UI \n");



  tbc = GUI_Toolbox (parent);

  box0 = GUI_Vbox (tbc, 0);
  box1 = GUI_Hbox (box0, 0);
  box2 = GUI_Hbox (box0, 0);

  GUI_Label (box1, " GIS1 ", 2);

  GUI_Butt (box1, "Import Points", gis_CB1,    (void*)"opePtab", 0);
  GUI_Butt (box1, "Export Points", gis_CB1,    (void*)"savPtab", 0);
  GUI_Butt (box1, "Delete Points", gis_CB1,    (void*)"delPtab", 0);
  GUI_Tip  (" deletes active surface and ALL its points ");
  GUI_Butt (box1, " create Mesh ", gis_CB1,    (void*)"crMsh", 0);

  // create (empty) list for the point-surfaces
  wls = GUI_OptMen (box1, NULL, gis_Lst_CB, -30);
  GUI_Tip  (" change active surface ");

  // wcbdp = GUI_CkButt (box1, "disp.points", 1, NULL, NULL, 0);
  // wcbdi = GUI_CkButt (box1, "disp.info", 1, NULL, NULL, 0);

  wcbcp = GUI_CkButt (box1, "createPoint", 0, gis_CB1, (void*)"PtCre", 0);

  GUI_Butt (box1, "Exit", gis_CB1,    (void*)"Exit", 0);
  GUI_Butt (box1, "Help", gis_CB1,    (void*)"Help", 0);


  //----------------------------------------------------------------
  wlbas = GUI_Label (box2, " Surf.Nr A- ", 2);

  wlbap = GUI_Label (box2, " PointNr P- ", 2);

  wenvx = GUI_Entry (box2, " X: ", "", NULL, -90);
  wenvy = GUI_Entry (box2, " Y: ", "", NULL, -90);
  wenvz = GUI_Entry (box2, " Z: ", "", NULL, -90);

  GUI_Butt (box2, "Del.Pt.", gis_CB1,    (void*)"PtDel", 0);
  GUI_Butt (box2, "Mod.Pt.", gis_CB1,    (void*)"PtMod", 0);
  GUI_Butt (box2, "Add Pt.", gis_CB1,    (void*)"PtAdd", 0);

  GUI_Butt (box2, "Test", gis_CB1,    (void*)"fTest", 0);

  gtk_widget_show (tbc);

  return tbc;

}


//=====================================================================
  int gis_CB1 (GtkWidget *parent, void *data) {
//=====================================================================

  char *cp1, cbuf1[256];

  cp1 = (char*)data;

  printf("gis_CB1  /%s/\n",cp1);



  //-------------------------------------------------
  if(!strcmp(cp1, "Help")) {
    APP_Help (APP_NAME);



  //-------------------------------------------------
  } else if(!strcmp(cp1, "Exit")) {
    gCad_fini ();


  //-------------------------------------------------
  } else if(!strcmp(cp1, "opePtab")) {
    APP_Open ("Open Pointfile", (void*)gis_ptab_load);


  //-------------------------------------------------
  } else if(!strcmp(cp1, "savPtab")) {
    APP_Save ("Save Pointfile", "points.dat", (void*)gis_ptab_save);


  //-------------------------------------------------
  } else if(!strcmp(cp1, "delPtab")) {
    gis_PtabDel();


  //-------------------------------------------------
  } else if(!strcmp(cp1, "crMsh")) {
    gis_mesh ();


  //-------------------------------------------------
  } else if(!strcmp(cp1, "PtDel")) {
    gis_PtDel();


  //-------------------------------------------------
  } else if(!strcmp(cp1, "PtMod")) {
    gis_PtMod();


  //-------------------------------------------------
  } else if(!strcmp(cp1, "PtAdd")) {
    gis_PtAdd();


  //-------------------------------------------------
  } else if(!strcmp(cp1, "PtCre")) {
    if(GUI_CkButt_get (wcbcp) == 1) 
      TX_Print("indicate position or select cad-point or select gis-point");


  //-------------------------------------------------
  } else if(!strcmp(cp1, "fTest")) {
    gis_dump_surTab ();
    gis_dump_pTab ();



  //-------------------------------------------------
  } else {
    TX_Error("gis_CB1 E001 %s",cp1);
  }


  L_exit:
  return 0;

}


//================================================================
  int gis_mesh () {
//================================================================
// OFFEN:
// write "A# = MSH A<#_pTab>" into mainProg

  int     att, fNr, i1, ii, *iba, ibNr;
  long    dli, dbm, dbp;
  Mesh    ms1;
  Fac3    *fTab = NULL;


  printf("gis_mesh %d\n",ActSurNr);
  if(ActSurNr < 0) return 0;



  // get surTab-recordNr
  ii = gis_surInd_surNr (ActSurNr);
  if(ii < 0) return -1;


  // fix mesh-surf-Nr
  if(surTab[ii].indm > -1) {           // use existing
    dbm = surTab[ii].indm;

  } else {                             // create new surf
    ++surIndLast;
    dbm = surIndLast;
      printf(" next A-nr = %d\n",dbm);
    surTab[ii].modif = 1;
  }
    printf(" dbm=%d\n",dbm);




  // reload points > pTab if modfied
  gis_bload_pTab (TYP_FuncInit);
  printf("gis_mesh 1 %d\n",ActPtabNr);


  // update points-display
  gis_HideAllPoints ();
  gis_DrawPoints (ActPtab, ActPtabNr, ActSurNr, 0);
  surTab[ii].visi = 1;


  // get space for boundary
  iba = (int*)memspc101;
  ibNr = sizeof(memspc101) / sizeof(int);    // 4 byte

  // create Mesh
  MSH_msh1 (&ms1, iba, &ibNr, ActPtab, ActPtabNr);
  printf("gis_mesh 2 %d\n",ibNr);


  // save Mesh into binFile
  fTab = ms1.f->fTab;
  fNr = ms1.f->fNr;
  MSH_bsav_fTab (fTab, fNr, WC_modnr, dbm);
  // do not overwrite binFile in gis_bload_fTab
  ActFtabModFlag = 0;
  gis_bload_fTab (TYP_FuncEnd);
  printf("gis_mesh 3 %d\n",fNr);



  // update the boundary-display
  // if(surTab[ii].indm > -1) {
    dbp = surTab[ii].indp;
    // find dli of point-surf
    dli = DL_find_obj (Typ_SUR, dbp, -1L);
      printf(" dbp=%d dli=%d\n",dbp,dli);
    DL_SetInd (dli);             // modify (do not create new DL-Record)
    att = 5;                     // hier act Color uebernehmen !!!
    dli = DL_StoreObj (Typ_SUR, dbp, att);
    GL_DrawCvIpTab (&dli, att, iba, ibNr, ActPtab, 1);
  // }


  // update the mesh-display
  gis_msh_disp (1, ii, dbm);


  printf("gis_mesh 4\n");


  DL_Redraw ();

  return 0;

}


//================================================================
  int gis_msh_disp (int styl, int inds, long dbm) {
//================================================================
// styl: 0=ON=shade; 1=OFF=symbolic
// inds  index to surTab


  int     iFirst, att;
  long    dli;


  printf("gis_msh_disp styl=%d inds=%d dbm=%d\n",styl,inds,dbm);
// gis_dump_surTab ();


  // first time to create mesh ?
  if(surTab[inds].indm != dbm) {
    surTab[inds].indm = dbm;
    iFirst = 1;  // yes
  } else {
    iFirst = 0;  // no
  }


  // load points > pTab
  gis_bload_pTab (TYP_FuncInit);
  printf("gis_msh_disp 1 %d\n",ActPtabNr);



  // load Mesh surTab[inds].indm > fTab
  gis_bload_fTab (TYP_FuncInit);
  printf("gis_msh_disp 2 %d\n",ActFtabNr);


/*
    col = (ColRGB*)&GR_ObjTab[dli].iatt;
    col->vsym = 0;
*/

  // (Re-)Display mesh
  if(iFirst == 0) {
    // use existing dli - modify
    dli = DL_find_obj (Typ_SUR, dbm, -1L);
      printf(" dbm=%d dli=%d\n",dbm,dli);
    DL_SetInd (dli);  // modify (do not create new DL-Record)
  }




  att = 8;   // hier act Color uebernehmen !!!

  dli = DL_StoreObj (Typ_SUR, dbm, att);
  GL_Surf_Ini (&dli, (void*)&att);
  GR_DrawFtab (ActPtab, ActFtab, ActFtabNr, styl);
  GL_EndList ();



  return 0;

}
 

//================================================================
  int gis_disp_surNr () {
//================================================================
// display activeSurfNr & activePointNr

  char      cbuf[128];


  // display new activeSurfNr in Label
  if(ActSurNr >= 0) {
    sprintf(cbuf, " SurfNr. = A%d ",ActSurNr);
  } else {
    sprintf(cbuf, " SurfNr. = -- ");
  }
  GUI_Label_mod (wlbas, cbuf);


  // display new activePointNr in Label
  if(ActPtNr >= 0) {
    sprintf(cbuf, " PointNr. = P%d ",ActPtNr);
  } else {
    sprintf(cbuf, " PointNr. = -- ");
  }
  GUI_Label_mod (wlbap, cbuf);

  return 0;

}


//================================================================
  int gis_disp_coords (Point *p1) {
//================================================================

  char      cbuf[60];

  cbuf[0] = '\0';
  UTX_add_fl_u (cbuf, p1->x);
  GUI_Entry_set (wenvx, cbuf);
  cbuf[0] = '\0';
  UTX_add_fl_u (cbuf, p1->y);
  GUI_Entry_set (wenvy, cbuf);
  cbuf[0] = '\0';
  UTX_add_fl_u (cbuf, p1->z);
  GUI_Entry_set (wenvz, cbuf);

  return 0;

}


//================================================================
  int gis_sel_CB (int src, long dl_ind) {
//================================================================
// userselection-callback; CAD-Core reports selection

  int       i1, icp;
  long      dli, dbi;
  char      cbuf[128];
  Point     *pt1, p1;
  DL_Att    dla;

  printf("gis_sel_CB src=%d dl_ind=%d\n",src,dl_ind);

  icp  = GUI_CkButt_get (wcbcp);  // createPoint; 0=no, 1=yes

  if(dl_ind < 0) {     // indicate
    if(icp == 0) return 0;
    // get userCoords of indicated point
    p1 = GL_GetViewPos ();
    // create gis-point
    gis_disp_coords (&p1);
    gis_PtAdd ();
    return 0;
  }


  // get DL-record of selected object
  dla = DL_GetAtt(dl_ind);

  // cad-point: create gis-point from cad-point
  if(dla.typ == Typ_PT) {
    if(icp == 0) return 0;
    // get point from DB
    p1 = DB_GetPoint (dla.ind);
    // create gis-point
    gis_disp_coords (&p1);
    gis_PtAdd ();
    return 0;
  }



  if(dla.typ != Typ_APPOBJ) return 0;      // skip Basic-objects


  // decode Object
  ActPtDli = dl_ind;
  ActSurNr = dla.ind;      // Surf.Nr
  ActPtNr  = dla.iatt;     // PointNr


  // display activeSurfNr & activePointNr
  gis_disp_surNr ();


  // display PointCoords in wenvx/y/z
  gis_disp_coords (&ActPtab[ActPtNr]);


  // create CAD-point
  if(icp == 0) return 0;

  i1 = DB_QueryLastUsed (Typ_PT, 1);
  ++i1;
  sprintf(cbuf, "P%d=",i1);
  AP_obj_add_pt(cbuf, &ActPtab[ActPtNr]);
  ED_add_Line (cbuf);
  DL_Redraw ();                 // Redraw DispList

  return 0;

}


//===================================================================
  int gis_bload_pTab (int mode) {
//===================================================================
// load all point of surface ActSurNr into ActPtab[ActPtabNr].
// mode=TYP_FuncInit   Load new
// mode=TYP_FuncEnd    unload.

  int   i1, ii;

  static int oldSur = -1;


  printf("gis_bload_pTab mode=%d sur=%d pNr=%d\n",mode,ActSurNr,ActPtabNr);
  printf("               oldSur=%d flg=%d\n",oldSur,ActPtabModFlag);



  // TYP_FuncEnd:  save points into bin.File
  if(mode == TYP_FuncEnd) {
    if(ActPtab) {
      if(ActPtabModFlag) {
        ii = gis_surInd_surNr (ActSurNr);
        surTab[ii].modif = 1;
        gis_bsave_pTab ();
      }
      free (ActPtab);
      ActPtab = NULL;
    }
    return 0;
  }

 
  // TYP_FuncInit = Load.


  if(ActPtab == NULL) goto L_new_load;
  if(oldSur != ActSurNr) goto L_load;
  if(ActPtabModFlag == 0) return 0;


  L_load:

  if(ActPtabModFlag) gis_bsave_pTab ();
  free(ActPtab);


  L_new_load:
  // load pTab from bin.file; mallocs ActPtab !
  i1 = MSH_bload_pTab (&ActPtab, &ActPtabNr, -1, ActSurNr);
  if(i1 < 0) return -1;
    printf(" ActPtabNr=%d\n",ActPtabNr);

  oldSur = ActSurNr;
  ActPtabModFlag = 0;    // unmodified

  // dump pTab
  // for(i1=0; i1<ActPtabNr; ++i1)
  // UT3D_stru_dump (Typ_PT,&ActPtab[i1],"P[%d]",i1);

  return 0;

}

//================================================================
  int gis_bsave_fTab (int dbSurM) {
//================================================================
 
  MSH_bsav_fTab (ActFtab, ActFtabNr, -1, dbSurM);

  return 0;

}


//================================================================
  int gis_bload_fTab (int mode) {
//================================================================

  int    ii, dbSurM;

  static int oldSur = -1;


  printf("gis_bload_fTab %d\n",mode);


  ii = gis_surInd_surNr (ActSurNr);
  if(ii < 0) return -1;

  dbSurM = surTab[ii].indm;
    printf(" ActSurNr=%d ii=%d dbSurM=%d\n",ActSurNr,ii,dbSurM);


  // TYP_FuncEnd:  save Mesh into bin.File
  if(mode == TYP_FuncEnd) {
    if(ActFtab) {
      if(ActFtabModFlag) {
        surTab[ii].modif = 1;
        gis_bsave_fTab (dbSurM);
      }
      free (ActFtab);
      ActFtab = NULL;
    }
    return 0;
  }


  if(ActFtab == NULL) goto L_new_load;
  if(oldSur != dbSurM) goto L_load;
  if(ActFtabModFlag == 0) return 0;


  L_load:

  if(ActFtabModFlag) gis_bsave_fTab (dbSurM);

  free(ActFtab);


  L_new_load:
  // load pTab from bin.file; mallocs ActPtab !
  MSH_bload_fTab (&ActFtab, &ActFtabNr, -1, dbSurM);
    printf(" ActFtabNr=%d\n",ActFtabNr);

  oldSur = dbSurM;
  ActFtabModFlag = 0;    // unmodified


  return 0;

}


//================================================================
  int gis_bsave_pTab () {
//================================================================

  int   i1;


  printf("gis_bsave_pTab %d\n",ActPtabNr);
 
  // delete unused points
  for(i1=ActPtabNr-1; i1 >= 0; --i1) {
    if(ActPtab[i1].x != UT_VAL_MAX) continue;
    MEM_del_nrec (&ActPtabNr, ActPtab, i1, 1, sizeof(Point));
  }

  // save pTab
  MSH_bsav_pTab (ActPtab, ActPtabNr, -1, ActSurNr);


  return 0;

}

 
//=====================================================================
  int gis_lst_sup () {
//=====================================================================
// create list of point-surfaces from surTab

  int    i1, sl;
  char   *pBuf[MAX_SURPTAB], cbuf1[40], *cBuf2;


  cBuf2 = memspc55;


  // Create a List of Pointers for a list of words.
  for(i1=0; i1<surTabNr;  ++i1) {
    sprintf(cbuf1, " A%d ", surTab[i1].indp);
    sl = strlen(cbuf1);
    strcpy(cBuf2, cbuf1);
    cBuf2[sl] = '\0';
    pBuf[i1] = cBuf2;
    cBuf2 += sl+1;
  }

  pBuf[surTabNr] = NULL; // terminat List



  // GUI_OptMen (parent, pBuf, gis_Lst_CB, 0);

  GUI_OptMenChg (wls, pBuf, gis_Lst_CB);




  return 0;

}


//===================================================================
  int gis_Lst_CB (GtkWidget *parent, void *data) {
//===================================================================
// userselection of listEntry

  int   lNr;

  lNr = (int)data;

  printf("gis_Lst_CB A%d\n",surTab[lNr].indp);


  /// activate surface surTab[lNr]
  gis_surPt_chg (lNr);



  return 0;

}


//================================================================
  int gis_init_pl () {
//================================================================
// fill surTab-List from ModelCodes


/*
  int      i1, grpNr;
  ObjGrp   grp[MAX_SURPTAB];
  ObjGX    flt[1];
  flt[0].typ = Typ_SURPTAB;  // filter all surfaces
  grpNr = MAX_SURPTAB;
  i1 = DB_Filter__ (&grp, &grpNr, &flt, 1);
  // fehlerhaft; da Objekte in der DB nicht geloescht werden !
*/


  int   irc, typ, ii, ind1, ind2, lNr, lLen;
  char  *lSta, *p1;


  printf("---------- gis_init_pl ----------\n");

  lSta = NULL;

  surTabNr = 0;
  for(;;) {
    // get next surface
    irc = APED_nxt_def_typ (&lSta, &lNr, &lLen, Typ_SUR);
    if(irc < 0) break;

      // printf(" nxt ln |");UTX_dump_cnl (lSta, 40);printf("|\n");
      

    // check for PTAB-SUR; next word after '=' must be PTAB
    p1 = UTX_find_Del1 (lSta);
    while(*p1 == ' ') ++p1;
    if(*p1 != '=') continue;

    irc = APT_decode_obj (&typ, &ind1, lSta);
    if(irc != 0) continue;
      // printf(" typ=%d ind1=%d\n",typ,ind1);

    // find next delimiter
    ++p1;
    while(*p1 == ' ') ++p1;

    if(!strncmp(p1, "PTAB", 4)) {
      surTab[surTabNr].indp   = ind1;           // DB-Index.
      surTab[surTabNr].indm   = -1;
      surTab[surTabNr].visi  = 0;               // not yet visible
      surTab[surTabNr].exi_p = 1;               // 1=exists in Model
      surTab[surTabNr].exi_m = 0;               // 0=does not exists in Model
      surTab[surTabNr].modif = 0;               // 0=unmodfied
      ++surTabNr;
    }

    if(!strncmp(p1, "MSH", 3)) {
      // find db-ind of pointSurface
      p1 += 4;
      while(*p1 == ' ') ++p1;
      irc = APT_decode_obj (&typ, &ind2, p1);
        // printf(" p1=|%s| %d %d\n",p1,typ,ind2);
      if(irc != 0) continue;
      if(typ != Typ_SUR) {printf("gis_init_pl E001\n"); continue;}
      // find pointSurf with ind ind1; set mesh-ind to ind2
      ii = gis_surInd_surNr (ind2);
      if(ii < 0) {printf("gis_init_pl E002\n"); continue;}
      surTab[ii].indm = ind1;
      surTab[ii].exi_m = 1;
        printf(" surTab[%d].indm = %d\n",ii,ind2);
    }

  }


  printf("ex gis_init_pl %d\n",surTabNr);


  return 0;

}


//================================================================
  int gis_UI_getPt (Point *pt1) {
//================================================================

  char    *p1, *p2;


  p1 = GUI_Entry_get (wenvx);
  if(strlen(p1) < 1) goto L_err1;
  pt1->x = strtod (p1, &p2);

  p1 = GUI_Entry_get (wenvy);
  if(strlen(p1) < 1) goto L_err1;
  pt1->y = strtod (p1, &p2);

  p1 = GUI_Entry_get (wenvz);
  if(strlen(p1) < 1) goto L_err1;
  pt1->z = strtod (p1, &p2);

  UT3D_stru_dump(Typ_PT, pt1, "ex gis_UI_getPt:");

  return 0;

  L_err1:
  TX_Print("***** coordinate - inputfield empty *****");
  return -1;

}


//================================================================
  int gis_surInd_surNr (int surNr) {
//================================================================
// get index into surTab from DB-index of pointSurface

  int surInd;

  for(surInd=0; surInd<surTabNr;  ++surInd) {
    if(surTab[surInd].indp == surNr) return surInd;
  }

  return -1;

}

 
//================================================================
  int gis_PtabDel () {
//================================================================
// delete (all Points from) surface ActSurNr

  int   irc, i1, surInd, lNr;
  long  dli, dbi;
  char  *lPos, oNam[32];

  printf("gis_PtabDel %d\n",ActSurNr);

  gis_dump_surTab ();


  // find index surInd of ActSurNr in surTab
  surInd = gis_surInd_surNr (ActSurNr);
  if(surInd < 0) return -1;
    printf(" surInd = %d\n",surInd);




  //----------------------------------------------------------------
  // delete mesh
  dbi = surTab[surInd].indm;
  if(dbi < 0) goto L_p;
    printf(" dbi-m=%d\n",dbi);


  if(surTab[surInd].indm < 0) goto L_p;

  // // delete bin.Mesh
  // MSH_bdel_fTab (WC_modnr, dbi);

  if(surTab[surInd].exi_m == 0) {
    // delete mesh-display
    dli = DL_find_obj (Typ_SUR, dbi, -1L);
     // printf(" del %d %d\n",dbi,dli);
    GL_Del0 (dli);

  } else {
    // delete Mesh from model (deletes also its display)
    gis_mdl_del_msh (surInd);
  }




  //----------------------------------------------------------------
  // delete all points from display
  L_p:
  if(surTab[surInd].visi == 1)
    gis_HidePoints (surTab[surInd].indp);


  // delete pTab-Code from model (del A<surInd> - Line)
  if(surTab[surInd].exi_p == 0) {
    dbi = surTab[surInd].indp;
    if(dbi < 0) goto L_9;
      printf(" dbi-p=%d\n",dbi);
    // delete mesh-display
    dli = DL_find_obj (Typ_SUR, dbi, -1L);
     // printf(" del %d %d\n",dbi,dli);
    GL_Del0 (dli);

  } else {
    gis_mdl_del_pTab (surInd);
  }


  // // delete bin. pointFile
  // MSH_bdel_pTab (WC_modnr, ActSurNr);



  //----------------------------------------------------------------
  L_9:
  // delete record # surInd from surTab
  MEM_del_nrec (&surTabNr, surTab, surInd, 1, sizeof(typ_sur_ptab));

  // update surf-list
  gis_lst_sup ();

  // activate last surf
  gis_surPt_chg (surTabNr-1);

  DL_Redraw ();

  return 0;

}

 
//================================================================
  int gis_PtDel () {
//================================================================
// delete Point Nr ActPtNr from surface ActSurNr

  printf("gis_PtDel %d %d %d\n",ActSurNr,ActPtNr,ActPtDli);


  if(ActPtNr < 0) {
    TX_Print(" select point to delete ..");
    return 0;
  }
  if(ActPtNr >= ActPtabNr) { TX_Error("gis_PtDelE001"); return -1; }


  // // load points > pTab
  // gis_bload_pTab (TYP_FuncInit);


  // delete point ActPtNr from Display
  GL_Del0 (ActPtDli);
  // DL_unvis_set (ActPtDli, 1);
  DL_Redraw ();                 // Redraw DispList


  // delete point ActPtNr in pTab
  ActPtab[ActPtNr].x = UT_VAL_MAX;


  // update Labels; ActPtNr-display
  ActPtNr = -1;
  gis_disp_surNr ();


  ActPtabModFlag = 1;
  ActFtabModFlag = 1;
 
  return 0;

}


//================================================================
  int gis_PtMod () {
//================================================================
// modify Point Nr ActPtNr from surface ActSurNr

  int    i1;
  Point  pt1;


  printf("gis_PtMod %d %d %d\n",ActSurNr,ActPtNr,ActPtDli);


  if(ActPtNr < 0) {
    TX_Print(" select point to modify ..");
    return 0;
  }


  // get point from UI
  i1 = gis_UI_getPt (&pt1);
  if(i1 < 0) return -1;



  // // load points > pTab
  // gis_bload_pTab (TYP_FuncInit);


  // modify point ActPtNr from Display
  DL_SetInd (ActPtDli);   // modify DL
  gis_DrawPoints (&pt1, 1, ActSurNr, ActPtNr);
  DL_Redraw ();                 // Redraw DispList


  // modify point ActPtNr in pTab
  ActPtab[ActPtNr] = pt1;

  ActPtabModFlag = 1;
  ActFtabModFlag = 1;

  return 0;

}


//================================================================
  int gis_PtAdd () {
//================================================================
// add Point to surface ActSurNr

  int    i1;
  Point  pt1;


  printf("gis_PtAdd %d %d\n",ActSurNr,ActPtabNr);

  // get point from UI
  i1 = gis_UI_getPt (&pt1);
  if(i1 < 0) return -1;


  // // load points > pTab
  // gis_bload_pTab (TYP_FuncInit);


  // get a free pos in pTab
  ActPtNr = gis_freePos_pTab ();
  if(ActPtNr < 0) return -1;


  // unhilite active point
  GL_UnHiliObj (ActPtDli);


  // create point ActPtabNr from Display
  // hier fehlt noch PunktIndexOffset !
  gis_DrawPoints (&pt1, 1, ActSurNr, ActPtNr);


  // hilite last created object
  GL_HiliObj (ActPtDli);
  DL_Redraw ();                 // Redraw DispList


  // save point in pTab
  ActPtab[ActPtNr] = pt1;

  // update ActPtNr-display
  gis_disp_surNr ();


  ActPtabModFlag = 1;
  ActFtabModFlag = 1;

  return 0;

}


//================================================================
  int gis_freePos_pTab () {
//================================================================
// get next free Position in pTab

   int i1, ii;


  for(i1=0; i1<ActPtabNr; ++i1) {
    if(ActPtab[i1].x == UT_VAL_MAX) goto L_exit;
  }


  // no free pos; realloc.
  ii = ActPtabNr + (ActPtabNr / 3);
  ActPtab = (Point*)realloc(ActPtab, ii * sizeof(Point));
  if(ActPtab == NULL) {TX_Error("gis_freePos_pTab EOM"); return -1;}

  // new points are unused
  for(i1=ActPtabNr; i1<ii; ++i1) ActPtab[i1].x = UT_VAL_MAX;

  i1 = ActPtabNr;
  ActPtabNr = ii;

  L_exit:
  printf("ex gis_freePos_pTab %d\n",i1);
  return i1;

}
 

//================================================================
  int gis_ptab_save (char *fnam, char *dirNam) {
//================================================================
// Export points into Ascii-file


  int   i1;
  char  cbuf[256];
  FILE  *fpo;

  if(fnam == NULL) return 0;


  printf("gis_ptab_save |%s|%s|\n",dirNam,fnam);


  // unload active surf
  gis_bload_pTab (TYP_FuncEnd);

  // load points > pTab
  gis_bload_pTab (TYP_FuncInit);


  if(ActPtabNr < 1) {
    TX_Print(" pTab not active ..");
    return 0;
  }


  sprintf(cbuf,"%s%s",dirNam,fnam);
  if((fpo=fopen(cbuf,"w")) == NULL) {
    TX_Print("gis_ptab_save E001 %s",fnam);
    return -1;
  }


  for(i1=0; i1<ActPtabNr; ++i1) {
    // printf(" %d %f %f %f\n",i1,
            // ActPtab[i1].x,ActPtab[i1].y,ActPtab[i1].z);
    fprintf(fpo,"P%d %f %f %f\n",i1,
            ActPtab[i1].x,ActPtab[i1].y,ActPtab[i1].z);
  }

  fclose(fpo);

  return 0;

}


//================================================================
  int gis_surPt_chg (int surInd) {
//================================================================
// disactivate surface surTab[actInd]
// activate surface surTab[surInd]

  int        irc, actInd;
  long       dli, dbi;
  ColRGB     *col;



  actInd = gis_surInd_surNr (ActSurNr);
  // change actInd > surInd


  printf("gis_surPt_chg %d act=%d\n",surInd,actInd);


  // no change: exit ...
  if(surInd == actInd) return 0;
  // if(ActSurNr == surTab[surInd].indp) return 0;


  // if no surf exists
  if(surInd < 0) {
    ActSurNr = -1;
    ActPtNr = -1;
    gis_disp_surNr ();
    return 0;
  }


  // unload active surf-points
  gis_bload_pTab (TYP_FuncEnd);


  // delete display of all points
  gis_HideAllPoints ();



  //----------------------------------------------------------------
  // change display of active mesh to shaded
  if(actInd < 0) goto L_1;

  dbi = surTab[actInd].indm;
  if(dbi < 1) goto L_1;
    printf(" m-act-dbi=%d\n",dbi);

  // if mesh already exists, just modify ..
  if(surTab[actInd].exi_m > 0) {
    // find dli of A<dbi>
    dli = DL_find_obj (Typ_SUR, dbi, -1L);
    if(dli < 0) goto L_1;
    GA_sStyl__ (dli, 1, Typ_SUR, dbi);    // change attrib & redraw


  } else {   // modify Attribut of temporary Element dli
    gis_msh_disp (0, actInd, dbi);
  }



  //----------------------------------------------------------------
  L_1:
  // load pTab from bin.file for surf. ActSurNr > ActPtab
  ActSurNr = surTab[surInd].indp;
  irc = gis_bload_pTab (TYP_FuncInit); // ActPtab[ActPtabNr]
  if(irc < 0) return -1;


  // show points
  gis_DrawPoints (ActPtab, ActPtabNr, ActSurNr, 0);
  surTab[surInd].visi = 1;



  //----------------------------------------------------------------
  // change display of new Mesh to symbolic
  dbi = surTab[surInd].indm;
  if(dbi < 1) goto L_2;
    printf(" m-new-dbi=%d\n",dbi);

    printf(" dbi=%d  dli=%d\n",dbi,dli);

  // if mesh already exists, just modify ..
  if(surTab[surInd].exi_m > 0) {
    // find dli of A<dbi>
    dli = DL_find_obj (Typ_SUR, dbi, -1L);
    if(dli < 0) goto L_2;
    GA_sStyl__ (dli, 2, Typ_SUR, dbi);    // change attrib & redraw


  } else {     // modify Attribut of temporary Element dli
    gis_msh_disp (1, surInd, dbi);
  }



  //----------------------------------------------------------------
  L_2:
  // activate actual surface in list
  GUI_OptMenActivate (wls, surInd);


  // display new activepointTab in Label
  ActPtNr = -1;
  gis_disp_surNr ();


  actInd = surInd;

  return 0;

};


//================================================================
  int gis_ptab_load (char *fnam, char *dirNam) {
//================================================================
// user has pointfile selected;
// see also AP_ImportXML

  int       irc, i1, ptNr;
  long      surNr;
  char      cbuf[256], attrib[2];
  Point     *pTab;


  if(fnam == NULL) return 0;

  strcpy(cbuf, dirNam);
  strcat(cbuf, fnam);


  printf("gis_ptab_load |%s|\n",cbuf);


  // assign memspace > pTab
  pTab = (void*)memspc101;
  ptNr = sizeof(memspc101) / sizeof(Point);
    printf(" pNr=%d\n",ptNr);


  // load points > pTab
  irc = ptab_file_load1 (pTab, &ptNr, 0, cbuf);
  if(irc < 0) return -1;


  // surNr = find next free surface-nr (A)
  ++surIndLast;
  surNr = surIndLast;
    printf(" next A-nr = %d\n",surNr);


  // save points to file M#A#.ptab
  MSH_bsav_pTab (pTab, ptNr, WC_modnr, surNr);


  // add surf to surTab
  gis_surTab_add (surNr, 0);
  // ActSurNr = surNr;


  // add surf to List
  gis_lst_sup ();


  // activate new surface
  gis_surPt_chg (surTabNr-1);


  return 0;

}


//=======================================================================
  int gis_DrawPoints (Point *pTab,  int ptNr, int surNr, int iOff) {
//=======================================================================
// Input:
//   iOff  the nr of the first point
// DispList:
//   typ   ApplicationObject (Typ_APPOBJ = 20)
//   indp   SurfNr (A#)
//   iatt  pointNr (index in pTab)

// Wie soll Bezug auf die Application eingebracht werden ?
// Gis hat viele ApplicationObjectTypes ..

// see APT_disp_SymB GL_DrawSymB
// see APT_disp_TxtA GR_DrawTxtA GL_DrawTxtA

  int    i1, icol1, icol2;
  long   l1;
  char   info[20];


  printf("gis_DrawPoints %d %d\n",ptNr,surNr);

  icol1 = 0;
  icol2 = 1;

  for(i1=0; i1<ptNr; ++i1) {
    sprintf(info, "%d", i1+iOff);   // itoa
    l1 = DL_StoreObj (Typ_APPOBJ, surNr, i1+iOff);
    gis_DrawPoint (&l1, SYM_TRI_S, &pTab[i1], icol1, icol2, info);
  }
  ActPtDli = l1;

  return 0;

}



//================================================================
  int gis_DrawPoint (long *ind, int symTyp, Point *pt1,
                     int icol1, int icol2, char *info) {
//================================================================

  GLuint dlInd;


  // printf("gis_DrawPoint %d %d\n",*ind,symTyp);


   // den DL-Index (+ Offset) holen)
   dlInd = GL_fix_DL_ind (ind);
     // printf(" dlInd=%d\n",dlInd);
  
   glNewList (dlInd, GL_COMPILE);
  
   glDisable (GL_LIGHTING);
  
   // nur fuer das Rescale mit Feedback erforderlich:
   glBegin (GL_POINTS);
   glVertex3dv ((double*)pt1);
   glEnd ();

   // display symbol symTyp
   glColor3fv    (GL_col_tab[icol1]);
   glRasterPos3d (pt1->x, pt1->y, pt1->z); // notwendig zum activate color!
   glCallList ((GLuint)symTyp);

   // display info
   glColor3fv    (GL_col_tab[icol2]);
   glRasterPos3d (pt1->x, pt1->y, pt1->z);
   glCallLists (strlen(info), GL_UNSIGNED_BYTE, (GLubyte*)info);


   glEnable (GL_LIGHTING);

   glEndList ();

   // GR_Disp_pt (&pTab[i1], SYM_TRI_S, 0);
   // GR_Disp_txi (&pTab[i1], i1, 0);

     // printf("ex gis_DrawPoint dli=%d\n",dlInd);
 
  return 0;

}

//================================================================
  int gis_HideAllPoints () {
//================================================================
// remove diplay of all Points

  int   i1;

  printf("gis_HideAllPoints \n");
gis_dump_surTab ();

  for(i1=0; i1<surTabNr;  ++i1) {
      printf(" AllPoints-%d vis=%d mod=%d\n",i1,
              surTab[i1].visi,surTab[i1].modif);
    if(surTab[i1].visi == 1) {
      gis_HidePoints (surTab[i1].indp);
      surTab[i1].visi = 0;
    }
  }

gis_dump_surTab ();
  printf("ex gis_HideAllPoints \n");

  return 0;

}

 
//================================================================
  int gis_HidePoints (int surNr) {
//================================================================

  printf("gis_HidePoints sur=%d\n",surNr);

  GL_Del_om (Typ_APPOBJ, surNr, -1);   // del multiple objects from DL
  GL_Regen1 ();                        // reset DL
  

  return 0;

}

 
//=========================================================
  int gCad_fini () {
//=========================================================
// dll being unloaded - reset Input, kill all open windows

  int    i1, ii;
  long   dbi, dli;


  printf("gCad_fini GIS1\n");
gis_dump_surTab ();


  // write to Main-Infowindow ..
  TX_Print("...... gCad_fini GIS1");


  if(tbGis1) {


    ii = gis_surInd_surNr (ActSurNr);  // get sufTab-index of active surf
    if(ii < 0) goto L_mu;


    //----------------------------------------------------------------
    // restore meshDisplay of active surf


    // GA_sStyl__ ruft APT_Draw__; dazu muss ein DB-Objekt existieren !
    dbi = surTab[ii].indm;
    if(dbi < 1) goto L_mu;     // ok; mesh exists ...

    // find dli of A<dbi>
    dli = DL_find_obj (Typ_SUR, dbi, -1L);
    if(dli < 1) goto L_mu;

      printf(" dbi=%d  dli=%d\n",dbi,dli);

    // if Mesh did exist before:
    if(surTab[ii].exi_m == 1) {
      GA_sStyl__ (dli, 1, Typ_SUR, dbi);    // change attrib & redraw


    } else {   // - new Mesh; Mdl-Code is created.
      GL_Del0 (dli);                        // delete mesh-display
      GA_delRec (Typ_SUR, dbi);             // reset Attribute of new MeshSurf.
    }




    //----------------------------------------------------------------
    // restore pointsDisplay of active surf
    L_mu:

    // remove the Display of all points
    gis_HideAllPoints ();

    // unload mesh
    gis_bload_fTab (TYP_FuncEnd);

    // unload points
    gis_bload_pTab (TYP_FuncEnd);





    //----------------------------------------------------------------
    // create update / Modelcodes
    gis_mdl_upd ();


    // remove the toolbar tbGis1
    GUI_del_widged (tbGis1);
    tbGis1 = NULL;

  }


  // reset  select (give back to mainApp)
  AP_UserSelection_reset ();
  APP_Close ();

  return 0;

}


//================================================================
  int gis_mdl_del_msh (int surInd) {
//================================================================
// delete mesh-codeline in model

  int      irc, lNr, ll, surNr;
  char     *lPos, cbuf[64];

 
  // no mesh to delete - did not exist before
  if(surTab[surInd].exi_m == 0) return 0;

  // no mesh created
  if(surTab[surInd].indm < 0) return 0;

  surNr = surTab[surInd].indm;

    AP_name_typInd (cbuf, Typ_SUR, surNr);
    // search defineition-line
    irc = APED_search_defLn (&lPos, &lNr, &ll, cbuf, -1);
    printf(" irc=%d lNr=%d |%s|\n",irc,lNr,cbuf);
    if(irc < 0) goto L_cr;

    // delete line in model
    APED_undo (lNr, lPos, 1);

  L_cr:

  return 0;

}


//================================================================
  int gis_mdl_del_pTab (int surInd) {
//================================================================
// delete pTab from model (del A<surInd> - Line)

  int       irc, lNr, ll, surNr;
  char      *lPos, cbuf[64];

  
  // no mesh to delete - did not exist before
  if(surTab[surInd].exi_p == 0) return 0;

  // points deleted ..
  if(surTab[surInd].indp < 0) return 0;

    surNr = surTab[surInd].indp;

    AP_name_typInd (cbuf, Typ_SUR, surNr);
    // search defineition-line
    irc = APED_search_defLn (&lPos, &lNr, &ll, cbuf, -1);
    printf(" irc=%d lNr=%d |%s|\n",irc,lNr,cbuf);
    if(irc < 0) goto L_cr;

    // delete line in model
    APED_undo (lNr, lPos, 1);
    // UTF_comment_line (lNr, lPos, 1, "_");

  L_cr:

  return 0;

}


//================================================================
  int gis_mdl_upd () {
//================================================================
// create update / Modelcodes

  int       i1;
  char      cbuf[80];


  printf("gis_mdl_upd %d\n",surTabNr);
    gis_dump_surTab ();


  for(i1=0; i1<surTabNr;  ++i1) {

    // if(surTab[i1].modif == 0) continue;


    //----------------------------------------------------------------
    // add A1=PTAB "pointfilename" to model;
    // pointfilename is unused !
    // sprintf(cbuf, "A%d=PTAB \"%s\"",surNr,fnam);
    if(surTab[i1].exi_p == 0) {    // did not exist before
      if(surTab[i1].indp > -1) {
        sprintf(cbuf, "A%d=PTAB",surTab[i1].indp);
          printf(" add |%s|\n",cbuf);
        ED_add_Line (cbuf);
      }
    }


    // add A1=MSH A<pointSurfNr>   to model;
    if(surTab[i1].exi_m == 0) {    // did not exist before
      if(surTab[i1].indm > -1) {
        sprintf(cbuf, "A%d=MSH A%d",surTab[i1].indm,surTab[i1].indp);
          printf(" add |%s|\n",cbuf);
        ED_add_Line (cbuf);
      }
    }

  }


  return 0;

}

 
//====================================================================
  int ptab_cad_load1 (Point *pTab, int *ptNr, Edg3 *lTab, int *lnNr) {
//====================================================================
// get points from DB &
// get breaklines from DB


  int      i1, ip1, ip2, ptMaxNr, lnMaxNr;
  long     dbi, dbMax;
  double   tol;
  Point    *pt1;
  Line     *ln1;

  printf("ptab_cad_load1 %d %d\n",*ptNr,*lnNr);


  ptMaxNr = *ptNr;
  lnMaxNr = *lnNr;

  // pTab ex DB-points laden.
  dbMax = DB_QueryLastUsed (Typ_PT, 0);
    printf(" PT-dbMax=%d\n",dbMax);

  // Punkte ex DB holen
  *ptNr = 0;
  for(dbi=1; dbi<=dbMax; ++dbi) {
    if(DB_QueryDef(Typ_PT, dbi) < 0) continue;
    pt1 = DB_get_PT (dbi);
    ip1 = ptab_ptAdd (pt1, pTab, ptNr, ptMaxNr, tol);
    if(ip1 < 0) break;
  }



  // Linien (breaklines) ex Db holen ...
  // neue Punkt an pTab anhaengen ..
  // ptMaxNr = *ptNr;
  dbMax = DB_QueryLastUsed (Typ_LN, 0);
    printf(" LN-dbMax=%d\n",dbMax);



  // Lines ex DB holen
  tol = UT_DISP_cv;
  i1 = 0;
  for(dbi=1; dbi<=dbMax; ++dbi) {
    if(DB_QueryDef(Typ_LN, dbi) < 0) continue;
    ln1 = DB_get_LN (dbi);
    // in pTab suchen, ob points already defined; ip1, ip2 setzen.
    // not yet defined: add to pTab.
    ip1 = ptab_ptAdd (&ln1->p1, pTab, ptNr, ptMaxNr, tol);
    ip2 = ptab_ptAdd (&ln1->p2, pTab, ptNr, ptMaxNr, tol);
    if(ip2 < 0) break;
    // add breakline
    if(i1 >= lnMaxNr) {TX_Print("ptab_cad_load1 E002"); break;}
    lTab[i1].i1 = ip1;
    lTab[i1].i2 = ip2;
    ++i1;
  }
  *lnNr = i1;



  // Test only ...........
    for(i1=0;i1<*ptNr;++i1) printf("p[%d]=%f %f %f\n",i1,
      pTab[i1].x,pTab[i1].y,pTab[i1].z);
    for(i1=0;i1<*lnNr;++i1) printf("l[%d]=%d %d\n",i1,
      lTab[i1].i1,lTab[i1].i2);


  return 0;

}


//==============================================================================
  int ptab_ptAdd (Point *pt1, Point *pTab, int *ptNr, int ptMaxNr, double tol) {
//==============================================================================
// check if pt1 is already in pTab; return its index,
// if not: add pt1 to pTab, return its index.
// RetCod = pointIndex or -1 (pTab.-Overfolw).
// used by ptab_cad_load1

  int   i1;

  UT3D_stru_dump(Typ_PT, pt1, "ptab_ptAdd:");
  printf(" ptNr=%d\n",*ptNr);

  // check if pt1 is already in pTab
  for(i1=0; i1<*ptNr; ++i1) {
    if(UT3D_comp2pt(pt1, &pTab[i1], tol) == 0) continue;
    return i1;   // gefunden ..
  }

  // add pt1 to pTab.
  i1 = *ptNr;
  if(i1 >= ptMaxNr) {TX_Print("ptab_ptAdd E001"); return -1;}
  pTab[i1] = *pt1;
  ++(*ptNr);


    GR_Disp_pt (&pTab[i1], SYM_TRI_S, 0);
    GR_Disp_txi (&pTab[i1], i1, 0);

  
  printf("ex ptab_ptAdd ptNr=%d\n",*ptNr);
  
  return i1;

}





//====================================================================
  int ptab_file_load1 (Point *pTab, int *ptNr, int mode, char *fNam) {
//====================================================================
// Punkte aus einer Ascii-Datei nach pTab einlesen.
// mode beschreibt das Dateiformat.
// Input:
//   ptNr   max nr of points on input !
//   mode   = Filetyp; 0 = N X Y Z C; 1 = X Y Z N C   UNUSED
//   fNam   incl. Filetyp

// Es wird getestet, ob erstes Wort numerisch.

// DefaultFileFormat:
// N=PointName, Text, zB "P001"; X,Y,Z=Coords; C=PointCode, Text, zB "Haus"
 
  int   i1, i2, ix, iy, iz, lNr, ptMaxNr;
  FILE  *fpi;
  char  cBuf[256], *cTab[6], *cp, *p1, cWd[80];


  printf("ptab_file_load1 |%s|\n",fNam);

  ptMaxNr = *ptNr;


  if ((fpi = fopen (fNam, "r")) == NULL) {
    TX_Print("FILE NOT FOUND |%s| in ptab_file_load1",fNam);
    return -1;
  }


  // fix the correlation-pointers
  // read first line & test first word.
  fgets (cBuf, 250, fpi);
/*
  cp = UTX_pos_skipDeli1 (cBuf); // skip leading blanks, tabs ..
  if(UTX_ck_num1(&p1, cp) == 0) {    // yes numer.
    ix = 0;
    iy = 1;
    iz = 2;
  } else {
    ix = 1;
    iy = 2;
    iz = 3;
  }
*/
  p1 = cBuf;
  i1 = -1;


  // search 1. word containing "."
  // - this is position of X-coord; Y-Coord and Z-Coord follows.

  // get next word
  L_nxt_wd:
  ++i1;
  cp = UTX_cp_word__ (cWd, p1);
    // printf(" wd[%d]=|%s|%s|\n",i1,cWd,cp);
  if(strchr(cWd, '.') == NULL) {
    p1 = UTX_pos_skipDeli1 (cp);
    if(i1 < 3) goto L_nxt_wd;
    TX_Print("FileFormat-Error %s",fNam);
    return -1;
  }


  // from extern:
  ix = i1;
  iy = ix + 1;
  iz = iy + 1;
    printf(" positions %d %d %d\n",ix,iy,iz);


  rewind(fpi);


  lNr = 0;
  while (!feof (fpi)) {
    if(fgets (cBuf, 250, fpi) == NULL) break;
    UTX_CleanCR (cBuf);    // cut off CR,LF 
      // printf("%s\n",cBuf);
    cp = cBuf;
    while (*cp == ' ') ++cp; // skip leading blanks
    if(strlen(cp) < 6) continue;
    // separate words ..
    cTab[0] = &cBuf[0];
    i2 = 0;

    // find starting-positions of first 6 words
    for(i1=0; i1<5; ++i1) {
      ++i2;
      // skip leading blanks
      cp = cTab[i1];
      // skip leading blanks
      while (*cp == ' ') ++cp;
      cTab[i2] = UTX_find_Del1 (cp);
      if(cTab[i2]) {*cTab[i2] = '\0'; ++cTab[i2];}   // separate with '\0'
    }
      // printf("|%s|%s|%s|%s|%s|\n",cTab[0],cTab[1],cTab[2],cTab[3],cTab[4]);

    // distribute; load Coords according to ix,iy,iz;
    pTab[lNr].x = strtod (cTab[ix], &cp);
    pTab[lNr].y = strtod (cTab[iy], &cp);
    pTab[lNr].z = strtod (cTab[iz], &cp);
      // UT3D_stru_dump (Typ_PT, &pTab[lNr], "%s:",cTab[0]);
      // GR_Disp_pt (&pTab[lNr], SYM_TRI_S, 0);
      // GR_Disp_txi (&pTab[lNr], lNr, 0);

    // proceed to next line ...
    ++lNr;
    if(lNr > ptMaxNr) {
      TX_Error("ptab_file_load1 EOM %d",ptMaxNr);
      return -1;
    }
  }

  fclose(fpi);

  *ptNr = lNr;
    printf("ex ptab_file_load1 points=%d\n",lNr);
  return 0;

}


//================================================================
  int gis_surTab_add (long indp, int visi) {
//================================================================
// add new created pointSurf to surTab

  if(surTabNr >= MAX_SURPTAB) return -1;

  surTab[surTabNr].indp  = indp;
  surTab[surTabNr].indm  = -1,       // does not yet exist
  surTab[surTabNr].visi  = visi;
  surTab[surTabNr].exi_p = 0;        // new created
  surTab[surTabNr].exi_m = 0;        // new created
  surTab[surTabNr].modif = 1;        // modified = yes

  ++surTabNr;

  return 0;

}
 

//================================================================
  int gis_dump_pTab () {
//================================================================
 
  int  i1;

  printf("========== gis_dump_pTab %d ===========\n",surTabNr);

  for(i1=0; i1<ActPtabNr;  ++i1) {
    UT3D_stru_dump (Typ_PT, &ActPtab[i1], "P[%d]",i1);
  }

  return 0;

}


//================================================================
  int gis_dump_surTab () {
//================================================================

  int  i1;

  printf("========== gis_dump_surTab %d ===========\n",surTabNr);
  printf(" active = %d PtabMod=%d FtabMod=%d\n",ActSurNr,
         ActPtabModFlag,ActFtabModFlag);


  for(i1=0; i1<surTabNr;  ++i1) {
    printf(" surTab[%d].indp=%d indm=%d vis=%d exp=%d exm=%d mod=%d\n",i1,
             surTab[i1].indp,
             surTab[i1].indm,
             surTab[i1].visi,
             surTab[i1].exi_p,
             surTab[i1].exi_m,
             surTab[i1].modif);

  }

  return 0;

}


//================  EOF  ==================================
